#include <stdio.h>   /* Standard input/output definitions */
#include <stdlib.h>
#include <string.h>  /* String function definitions */
#include <stdbool.h> /* Typedef for boolean */
#include <unistd.h>  /* UNIX standard function definitions */
#include <fcntl.h>   /* File control definitions */
#include <errno.h>   /* Error number definitions */
#include <termios.h> /* POSIX terminal control definitions */
#include <sys/ioctl.h>

#include "csp.h"

int fd = -1; /* File descriptor for the port */

static int CspVolume;					// Device volume setting
static int CspBarcodeRedundancy;		// Device barcode redundancy setting
static int CspContinuousScanning;		// Device continuous scan setting

static char CspTlBits[ 8 ];			// Device TL Bits
static char CspUploadCount[ 4 ];		// Device upload count
static char CspSignature[ 8 ];			// Device signature
static char CspUserId[ 8 ];			// Device User ID
static char CspSwVersion[ 8 ];			// Device software version
static char CspDeviceId[ 9 ];			// Device ID
static char CspBarData[ MAXSIZE ];		// uploaded barcode storage
static char ByteBuffer[ MAXSIZE ];		// temp storage of CSP device response

/*--------------------------------------------
 * Functions for communication with CyberPen
 *--------------------------------------------
 */

/*
 ** static char cspLrcCheck(char aLrcBytes[], int nMaxLength)
 *
 *  PARAMETERS:     aLrcBytes[] contains the bytes used for computing the LRC
 *                  nMaxLength determines how many bytes of aLrcBytes[] are used
 *                  to perform LRC checking.
 *
 *  DESCRIPTION:    Computes the Longitudinal Redundancy Check on a
 *                  string of specified length.
 *
 *  RETURNS:        The calculated LRC is returned.
 *
 */
static char cspLrcCheck( char LrcBytes[], int nMaxLength ) {
	int i;
	unsigned char chLrc = 0;        // initialize LRC value

	// compute the LRC for the entire string...
	for( i = 0; i < nMaxLength; i++ )
		chLrc ^= LrcBytes[ i ];

	return( chLrc );
}

static char cspGetc( void ) {
	char chRead;

	// make sure that a port has been selected...   
	if( fd > -1 ) {
		// a port is selected, can we retrieve any data from it?
		// try to read a character...
		usleep( 50 );
		if( read( fd, &chRead, 1 ) > 0 ) {
			return( chRead );
		}
		// let other tasks run while we wait for a character...
	}
   // if we make it this far, no character was read. So...  
   return( COMMUNICATIONS_ERROR );
}

static char cspSendCommand( char *Command, int nMaxLength ) {
	int i, sendcnt;
	char buffer[ 255 ];

	// make sure we have a valid port selected...
	if( fd == -1 ) {
		return( COMMUNICATIONS_ERROR );
	}

	// delay before sending message
	usleep( 120 );

	// clear the receive queue...
	while( read( fd, &buffer, sizeof(buffer) ) > 0 );

	sendcnt = 0;
	do {
		sendcnt++;
		// send the command string...
		for( i = 0; i < nMaxLength; i++ ) {
			write( fd, &Command[ i ], 1 );
		}

		// wait for the reply to be processed, or MAXTIME...
		ByteBuffer[ 0 ] = cspGetc();

	} while( ( ByteBuffer[ 0 ] < 0 ) && ( sendcnt < MAXTIME ) );

	// if response not available, reply with error
	if ( ByteBuffer[ 0 ] < 0 ) {
		return( COMMUNICATIONS_ERROR );
	}

	if ( ByteBuffer[ 0 ] != NO_ERROR_ENCOUNTERED ) {
		return ( ByteBuffer[ 0 ] );
	} else {
		return ( STATUS_OK );
	}
}

char cspWakeup( void ) {
	char Commands[] = { INTERROGATE, STX, 0x00, 0x03 };
	int i;

	if( cspSendCommand( Commands, sizeof( Commands ) ) == STATUS_OK ) {
		// device answered with NO_ERROR_ENCOUNTERED so, we read STX next
		ByteBuffer[ 1 ] = cspGetc();

		// Protocol-Type
		ByteBuffer[ 2 ] = cspGetc();
		char ptype = ByteBuffer[ 2 ];

		// 00-Systemstatus ok   (other-Low Battery Indication)
		ByteBuffer[ 3 ] = cspGetc();
		char state = ByteBuffer[ 3 ];

		// 8 Bytes User ID
		for( i = 4; i < 4 + 8; i++ ) {
			ByteBuffer[ i ] = cspGetc();
			CspUserId[ i - 4 ] = ByteBuffer[ i ];
		}

		// 8 Bytes Device Firmware Version
		for( i = 12; i < 12 + 8; i++ ) {
			ByteBuffer[ i ] = cspGetc();
			CspSwVersion[ i - 12 ] = ByteBuffer[ i ];
		}

		// 00-Null (message terminator)
		if( cspGetc() == 0 ) {
			ByteBuffer[ 20 ] = 0;
			// Compare LRC-Checksum
			if( cspLrcCheck( ByteBuffer, 20 ) == cspGetc() ) {
				// save values to variables
				
				printf( "Values:\nPType: %i\nSystemstatus: %i\nUserID: %i\nDFV: %s\n", ptype, state, CspUserId, CspSwVersion );
				
				return( STATUS_OK );
			} else {
				return( LRC_FAILED );
			}
		} else {
			return( COMMUNICATIONS_ERROR );
		}
	} else {
		return( COMMUNICATIONS_ERROR );
	}
}

char cspClear( void ) {
	char Commands[] = { STX, STX, 0x00, 0x00 };
	if( cspSendCommand( Commands, sizeof( Commands ) ) == STATUS_OK ) {
		// device answered with NO_ERROR_ENCOUNTERED so, we read STX next
		ByteBuffer[ 1 ] = cspGetc();
		// 00-Null
		ByteBuffer[ 2 ] = cspGetc();

		// Compare LRC-Checksum
		if( cspLrcCheck( ByteBuffer, 3 ) == cspGetc() ) {
			return( STATUS_OK );
		} else {
			return( LRC_FAILED );
		}
	} else {
		return( COMMUNICATIONS_ERROR );
	}
}

char cspPOff( void ) {
	char Commands[] = { POWER_DOWN, STX, 0x00, 0x07 };
	if( cspSendCommand( Commands, sizeof( Commands ) ) == STATUS_OK ) {
		// device answered with NO_ERROR_ENCOUNTERED so, we read STX next
		ByteBuffer[ 1 ] = cspGetc();
		// 00-Null
		ByteBuffer[ 2 ] = cspGetc();

		// Compare LRC-Checksum
		if( cspLrcCheck( ByteBuffer, 3 ) == cspGetc() ) {
			return( STATUS_OK );
		} else {
			return( LRC_FAILED );
		}
	} else {
		return( COMMUNICATIONS_ERROR );
	}
}

/*-------------------------------------------
 * Functions for setting up the serial port
 *-------------------------------------------
 */
int open_port( char *port ) {
	fd = open( port, O_RDWR | O_NOCTTY | O_NDELAY );
	if( fd == -1 ) {
		/*
		* Could not open the port.
		*/
		fprintf( stderr, "Unable to open Port %s - ", port );
		perror( "open_port" );

	} else {
		fcntl( fd, F_SETFL, 0 );
	}

	return( fd );

}

int init_port( void ) {
	struct termios options;
	int status = 0;
	status = ( TIOCM_RTS | TIOCM_DTR );

	/*
	 * Get the current options for the port...
	 */
	tcgetattr( fd, &options );

	/*
	 * Disable hardware- and software-flowcontrol
	 */
//	options.c_iflag &= ~(IGNBRK | BRKINT | ICRNL | INLCR | PARMRK | INPCK | ISTRIP | IXON);
	options.c_iflag &= ~( BRKINT | ICRNL | INLCR | PARMRK | INPCK | ISTRIP | IXON );

	/*
	 * Raw Input/Output Mode
	 */
	options.c_oflag = 0;
	options.c_lflag &= ~( ECHO | ECHONL | ICANON | IEXTEN | ISIG );

	/*
	 * Set Paritty Checking to 8n1
	 */
	options.c_cflag &= ~( CSIZE | PARENB );
	options.c_cflag |= CS8;

	options.c_cc[ VMIN ]  = 0;
	options.c_cc[ VTIME ] = 10;	// 1 sec timeout
	
	/*
	 * Set the baud rates to 2400...
	 */
	cfsetispeed( &options, B2400 );
	cfsetospeed( &options, B2400 );

	/*
	* Set the new options for the port...
	*/
	tcsetattr( fd, TCSANOW, &options );

	/*
	 * Set DTS and RTS to high, so dockingstation
	 * gets powered up
	 */
	ioctl( fd, TIOCMBIS, &status );

	return( 0 );
}

int close_port( void ) {
	close( fd );
	return( 0 );
}
